@@ -1,16 +1,17 @@
 from datetime import datetime
+from re import template
 from RupineHeroku.rupine_db import herokuDbAccess
 from psycopg2 import sql
 import json
 
-def POST(connection, schema, tableName:str, data:dict, onConflict:bool=False, uniqueColumnNameForConflict:str='id'):
+def POST(connection, schema, tableName:str, data:dict, onConflict:bool=False, uniqueColumnNamesForConflict:str='id'):
     # TODO: Check if any column is not nullable that does not appear in data. Return ERROR in this case
     data['created_at'] = int(datetime.now().timestamp())
     data['modified_at'] = int(datetime.now().timestamp())
     columns = data.keys()
     onConflictString = ''
     if onConflict:
-        onConflictString = 'ON CONFLICT ({}) DO NOTHING'.format(uniqueColumnNameForConflict)
+        onConflictString = 'ON CONFLICT ({}) DO NOTHING'.format(uniqueColumnNamesForConflict)
     queryString = "INSERT INTO {{}}.{} ({}) VALUES ({}) {};".format(tableName,', '.join(columns),','.join(['%s']*len(columns)),onConflictString)
 
     params = []
@@ -24,6 +25,49 @@ def POST(connection, schema, tableName:str, data:dict, onConflict:bool=False, un
     result = herokuDbAccess.insertDataIntoDatabase(query, params, connection)    
     return result
 
+def POST_BULK(connection, schema, tableName:str, data:list, byCopy = True, onConflict:bool=False, uniqueColumnNamesForConflict:str='id'):
+    '''
+    data must be list of diictionaries. Key is column name according to database!
+    '''
+    query = sql.SQL('SELECT column_name,data_type FROM information_schema.columns WHERE table_schema = %s AND table_name = %s')
+    res = herokuDbAccess.fetchDataInDatabase(query, [schema,tableName], connection)
+    columns = []
+    template = []
+    for row in res:
+        columns.append(row[0])
+        if row[1] == 'json':
+            template.append('%s::json')
+        else:
+            template.append('%s')
+    template = '(' + ','.join(template) + ')' 
+
+    newData = []
+    for item in data:
+        item['created_at'] = int(datetime.now().timestamp())
+        item['modified_at'] = int(datetime.now().timestamp())
+        # TODO: Check if any column is not nullable that does not appear in data. Return ERROR in this case
+        newItem = []
+        for column in columns:
+            if column in item:
+                if type(item[column]) == dict:
+                    newItem.append(json.dumps(item[column]))
+                else:
+                    newItem.append(item[column])
+            else:
+                newItem.append(None)
+        newData.append(newItem)
+
+    if byCopy and not onConflict:
+        result = herokuDbAccess.insertBulkDataIntoDatabaseByCopyManager('.'.join([schema,tableName]),columns,newData,connection)  
+    else:
+        onConflictString = ''
+        if onConflict:
+            onConflictString = 'ON CONFLICT ({}) DO NOTHING'.format(uniqueColumnNamesForConflict)
+        queryString = "INSERT INTO {{}}.{} ({}) VALUES %s {};".format(tableName,', '.join(columns),onConflictString)
+        query = sql.SQL(queryString).format(sql.Identifier(schema))
+        result = herokuDbAccess.insertBulkDataIntoDatabase(query,template,newData,connection)
+    return result
+
 def PUT(connection, schema, updates:dict, tableName:str, conditions:dict={}):
     updates['modified_at'] = int(datetime.now().timestamp())
     sqlTemplateEqual = "{} = %s"
@@ -33,7 +77,11 @@ def PUT(connection, schema, updates:dict, tableName:str, conditions:dict={}):
     params = []
     for key in updates.keys():
         setArray.append(sqlTemplateEqual.format(key))
-        params.append(updates[key])
+        if type(updates[key]) == dict:
+            params.append(json.dumps(updates[key]))
+            
+        else:
+            params.append(updates[key])
     
     for key in conditions.keys():
         if type(conditions[key]) == list:
@@ -52,6 +100,56 @@ def PUT(connection, schema, updates:dict, tableName:str, conditions:dict={}):
     herokuDbAccess.insertDataIntoDatabase(query, params, connection)    
     return None
 
+def PUT_BULK(connection, schema, updates:list, tableName:str, conditionColumns:list=['id']):
+    '''
+    updates must be list of dictionaries (with equal keys!). Key is column name according to database!
+    conditionColumns must be in updates!
+    '''
+    query = sql.SQL('SELECT column_name,data_type FROM information_schema.columns WHERE table_schema = %s AND table_name = %s')
+    res = herokuDbAccess.fetchDataInDatabase(query, [schema,tableName], connection)
+    columns = ['modified_at']
+    for row in res:
+        if row[0] in updates[0]:
+            columns.append(row[0])
+
+
+    newData = []
+    for item in updates:
+        item['modified_at'] = int(datetime.now().timestamp())
+        # TODO: Check if any column is not nullable that does not appear in data. Return ERROR in this case
+        newItem = []
+        for column in columns:
+            if column in item:
+                if type(item[column]) == dict:
+                    newItem.append(json.dumps(item[column]))
+                else:
+                    newItem.append(item[column])
+            else:
+                newItem.append(None)
+        newData.append(newItem)
+
+    setArray = []
+    conditionArray = []
+
+    for col in conditionColumns:
+        if col not in columns:
+            return {
+                'error': 1,
+                'msg': 'at least one columnn in conditionColumns is not in updates'
+            }
+        conditionArray.append("t.{} = tmp.{}".format(col,col))
+    for col in columns:
+        if col not in conditionColumns:
+            setArray.append("SET t.{} = tmp.{}".format(col,col))
+    
+    tmpTableName = schema + '_tmp_' + tableName
+    queryString = "UPDATE {{}}.{} t {} FROM {} tmp WHERE 1=1 AND {}".format(tableName,', '.join(setArray),tmpTableName,' AND '.join(conditionArray))
+
+    query = sql.SQL(queryString).format(sql.Identifier(schema))
+
+    result = herokuDbAccess.updatetBulkDataInDatabaseByCopyManager(schema,tableName,columns,newData,query,connection)   
+    return result
+
 def SELECT(connection, schema, columns:list, tableName:str, conditions:dict={}):
     if columns == [] or columns == ['*']:
         query = sql.SQL('SELECT column_name FROM information_schema.columns WHERE table_schema = %s AND table_name = %s')
@@ -123,10 +221,12 @@ def SELECT_FUNCTION(connection, schema,functionName,functionParameter:list,colum
         result.append(resultDict)
     return result
 
-# import os
-# from dotenv import load_dotenv
-# import herokuDbAccess as db
-# load_dotenv()
+import os
+from dotenv import load_dotenv
+import herokuDbAccess as db
+import time
+
+load_dotenv()
 
 # def convertToType(data,type):
 #     if data is None:
@@ -153,14 +253,65 @@ def SELECT_FUNCTION(connection, schema,functionName,functionParameter:list,colum
 #         'lp_pool_tokens':convertToType(res[15],float),
 #     }
 
-# if __name__ == '__main__':
-#     connection = db.connect(
-#         os.environ.get("HEROKU_DB_USER"),
-#         os.environ.get("HEROKU_DB_PW"),
-#         os.environ.get("HEROKU_DB_HOST"),
-#         os.environ.get("HEROKU_DB_PORT"),
-#         os.environ.get("HEROKU_DB_DATABASE")
-#     )
+if __name__ == '__main__':
+    connection = db.connect(
+        os.environ.get("HEROKU_GOD_USER"),
+        os.environ.get("HEROKU_GOD_PASSWORD"),
+        os.environ.get("HEROKU_HOST"),
+        os.environ.get("HEROKU_PORT"),
+        os.environ.get("HEROKU_DB")
+    )
+
+    start = time.time()
+    
+    for i in range(10,20):
+        data = {
+            'id': i,
+            'number':i+20
+        }
+        updates = {
+            'number':i+200
+        }
+        conditions = {
+            'id': i
+        }
+        #r= POST(connection,'dbdev','testbulk',data,True,'id')
+        r = PUT(connection,'dbdev',updates,'testbulk',conditions)
+        print(r)
+
+    end = time.time()
+    print(end - start)
+
+    start = time.time()
+    data = []
+    for i in range(1,10):
+        item = {
+            'id': i,
+            'number':i+200
+        }
+        data.append(item)
+    #r = POST_BULK(connection,'dbdev','testbulk',data,False,True,'id')
+    r = PUT_BULK(connection,'dbdev',data,'testbulk')
+    print(r)
+
+    end = time.time()
+    print(end - start)
+
+    
+
+
+
+#     data =  {'id': '30772bccd4b4e747b588e100c375242b8971d72c69871eb77fabfe478b7f43a6_4_-1_0', 'txid': '30772bccd4b4e747b588e100c375242b8971d72c69871eb77fabfe478b7f43a6', 'public_address': 'N/A', 'transaction_type': 'setgovheight', 'block_number': 2103358, 'block_timestamp': 1659194237, 'tx_order': 4, 'vin': None, 'vout': 0, 'value': 0, 
+#         'data': {
+#             'attributes': [
+#                 {'unknown': 1, 'type': 'p', 'typeId': 17, 'key': 'a', 'keyId': 1, 'value': 29270000},
+#                 {'unknown': 1, 'type': 'p', 'typeId': 101, 'key': 'b', 'keyId': 1, 'value': 29270000},
+#                 {'unknown': 1, 'type': 'p', 'typeId': 102, 'key': 'b', 'keyId': 1, 'value': 29270000}
+#             ],
+#         'height': 2105280, 'hex': None, 'name': 'setgovheight'}}
+#     updates = data.copy()
+#     updates.pop('id',None)
+#     PUT(connection,os.environ.get("ENVIRONMENT"),updates,'dfi_transaction',{'id':data['id']})
 #     #dfi_dex_calc_preparation_by_timestamp
 #     #dfi_dex_tx_get_all_by_timestamp
 #     SELECT_FUNCTION(connection,'dbdev','dfi_dex_tx_get_all_by_timestamp',[0, 1659183422])
