@@ -5,16 +5,21 @@ import pytz
 
 def postTaxTransaction(connection, schema, data):
 
-    query = sql.SQL("INSERT INTO {}.tax_transaction (tx_no,chain_id,chain,public_address,timestamp,block_number,transaction_hash,category,token,amount,usd_value,eur_value,fee_token,fee_amount,fee_usd_value,fee_eur_value) VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)").format(sql.Identifier(schema))
+    query = sql.SQL("INSERT INTO {}.tax_transaction (tx_no,chain_id,chain,location,location_type,public_address,timestamp,block_number,transaction_hash,interacted_with_contract,category_mapping_id,from_public_address,to_public_address,token,amount,usd_value,eur_value,fee_token,fee_amount,fee_usd_value,fee_eur_value) VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)").format(sql.Identifier(schema))
     params = (
         data['tx_no'],
         data['chain_id'],
         data['chain'],
+        data['location'],
+        data['location_type'],
         data['public_address'],
         data['timestamp'],
         data['block_number'],
         data['transaction_hash'],
-        data['category'],
+        data['interacted_with_contract'],
+        data['category_mapping_id'],
+        data['from_public_address'],
+        data['to_public_address'],
         data['token'],
         data['amount'],
         data['usd_value'],
@@ -61,14 +66,16 @@ def getTaxTransactionRewards(connection, schema, addresses:list=None, chain=None
         conditions = conditions + " AND t.amount < 0"
 
     query = sql.SQL("SELECT basis.* FROM ( \
-                        SELECT t.id,t.chain_id,t.chain,t.public_address,t.timestamp,t.block_number,t.transaction_hash,t.category,t.token,t.amount,t.usd_value,t.eur_value,t.fee_usd_value,t.fee_eur_value, \
+                        SELECT t.id,t.chain_id,t.chain,t.public_address,t.timestamp,t.block_number,t.transaction_hash,t.interacted_with_contract,c.category,t.token,t.amount,t.usd_value,t.eur_value,t.fee_usd_value,t.fee_eur_value, \
                         CASE WHEN c.sort_no IS NULL THEN '999' ELSE c.sort_no END AS sort_no \
-                        FROM {0}.tax_transaction t LEFT JOIN {0}.tax_category c ON t.category = c.category \
+                        FROM {0}.tax_transaction t LEFT JOIN {0}.tax_category_mapping cm ON t.category_mapping_id = cm.id \
+                        LEFT JOIN {0}.tax_category c ON cm.category_no = c.category_no \
                         WHERE 1=1 AND t.timestamp >= %%s %s \
                             UNION ALL	\
-                        SELECT t.id,t.chain_id,t.chain,t.public_address,t.timestamp,NULL AS block_number,NULL AS transaction_hash,t.category,t.token,t.amount,t.usd_value,t.eur_value,NULL AS fee_usd_value,NULL AS fee_eur_value, \
+                        SELECT t.id,t.chain_id,t.chain,t.public_address,t.timestamp,NULL AS block_number,NULL AS transaction_hash,NULL AS interacted_with_contract,c.category,t.token,t.amount,t.usd_value,t.eur_value,NULL AS fee_usd_value,NULL AS fee_eur_value, \
                         CASE WHEN c.sort_no IS NULL THEN '999' ELSE c.sort_no END AS sort_no \
-                        FROM {0}.tax_reward t LEFT JOIN {0}.tax_category c ON t.category = c.category \
+                        FROM {0}.tax_reward t LEFT JOIN {0}.tax_category_mapping cm ON t.category_mapping_id = cm.id \
+                        LEFT JOIN {0}.tax_category c ON cm.category_no = c.category_no \
                         WHERE 1=1 AND t.timestamp >= %%s %s \
                     ) basis ORDER BY basis.timestamp ASC, basis.sort_no ASC" % (conditions,conditions)).format(sql.Identifier(schema))
 
@@ -139,6 +146,27 @@ def getLatestWarehouseTxDate(connection, schema, token:str, account:str, chain=N
     return result[0]
     #return datetime(d.year,d.month,d.day, 0, 0, 0, tzinfo=pytz.timezone('UTC'))
 
+def getLatestTransactionRewardBlock(connection, schema, address:str, chain=None, chain_id=None):
+    conditions = ""
+    params = []
+    if chain is None and chain_id is None:
+        return []
+    elif chain is not None:
+        conditions = conditions + " AND chain = %s"
+        params.append(chain)
+    elif chain_id is not None:
+        conditions = conditions + " AND chain_id = %s"
+        params.append(chain_id)
+
+    query = sql.SQL("SELECT 'tax_transaction' AS table, MAX(block_number) as latest_block_number FROM {0}.tax_transaction \
+                    WHERE 1=1 AND public_address = %%s %s \
+                    UNION ALL \
+                    SELECT 'tax_reward' AS table, MAX(latest_block_number) as latest_block_number FROM {0}.tax_reward \
+                    WHERE 1=1 AND public_address = %%s %s" % (conditions,conditions)).format(sql.Identifier(schema))
+
+    result = herokuDbAccess.fetchDataInDatabase(query, [address,*params,address,*params], connection)  
+    return result
+
 def getLatestWarehouseDate(connection, schema, token:str, account:str, chain=None, chain_id=None):
     conditions = ""
     params = []
@@ -160,16 +188,18 @@ def getLatestWarehouseDate(connection, schema, token:str, account:str, chain=Non
     d = result[0][0]
     return datetime(d.year,d.month,d.day, 0, 0, 0, tzinfo=pytz.timezone('UTC'))
 
-
 def postTaxReward(connection, schema, data):
-    query = sql.SQL("INSERT INTO {}.tax_reward (tx_no,chain_id,chain,public_address,timestamp,category,token,amount,usd_value,eur_value) VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)").format(sql.Identifier(schema))   
+    query = sql.SQL("INSERT INTO {}.tax_reward (tx_no,chain_id,chain,location,location_type,public_address,timestamp,latest_block_number,category_mapping_id,token,amount,usd_value,eur_value) VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)").format(sql.Identifier(schema))   
     params = (
         data['tx_no'],
         data['chain_id'],
         data['chain'],
+        data['location'],
+        data['location_type'],
         data['public_address'],
         data['timestamp'],
-        data['category'],
+        data['latest_block_number'],
+        data['category_mapping_id'],
         data['token'],
         data['amount'],
         data['usd_value'],
@@ -229,7 +259,20 @@ def postTaxTrade(connection, schema, data):
     result = herokuDbAccess.insertDataIntoDatabase(query, params, connection)    
     return result
 
+def postCategoryMapping(connection, schema, category):
+    query = sql.SQL("SELECT id FROM {}.tax_category_mapping WHERE category_raw = %s;").format(sql.Identifier(schema))
+    params = [category]
+    result = herokuDbAccess.fetchDataInDatabase(query, params, connection)  
 
+    if len(result) > 0:
+        return result[0][0]
+
+    query = sql.SQL("INSERT INTO {}.tax_category_mapping (category_raw, category_no, is_edited) VALUES (%s,%s,%s) ON CONFLICT (category_raw) DO NOTHING RETURNING id;").format(sql.Identifier(schema))
+    params = [category,'-1','N']
+    result = herokuDbAccess.insertDataIntoDatabase(query, params, connection)
+    if result is not None:    
+        return result[0][0]
+    return result
 
 # import os
 # from dotenv import load_dotenv
@@ -244,7 +287,10 @@ def postTaxTrade(connection, schema, data):
 #         os.environ.get("HEROKU_DB_HOST"),
 #         os.environ.get("HEROKU_DB_PORT"),
 #         os.environ.get("HEROKU_DB_DATABASE")
-#     )
+#      )
+
+#     print(postCategoryMapping(connection,'dbdev','Exit staking wallet'))
+#     print(getLatestTransactionRewardBlock(connection,'stage','DFI','df1qgssyvcqld7cwxzvnajyt3xlqae67muxrqh06ct','DeFiChain'))
     # data = {
     #     'chain_id':1,
     #     'chain':'ETH',
